#!/usr/bin/env python

"""
    python davserver
    Copyright (C) 1999 Christian Scholz (ruebe@aachen.heimat.de)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""


from Ft.Xml import Domlette

import sys
import string
import urlparse
import urllib
from StringIO import StringIO

import utils
from constants import COLLECTION, OBJECT, DAV_PROPS, RT_ALLPROP, RT_PROPNAME, RT_PROP
from errors import *

DAV_NS = "DAV:"

class PROPFIND:
    """ parse a propfind xml element and extract props

    It will set the following instance vars:

    request_class   : ALLPROP | PROPNAME | PROP
    proplist    : list of properties
    nsmap       : map of namespaces

    The list of properties will contain tuples of the form
    (element name, ns_prefix, ns_uri)


    """


    def __init__(self,uri,dataclass,depth):
        self.request_type=None
    self.nsmap={}
    self.proplist={}
    self.default_ns=None
    self.__dataclass=dataclass
    self.__depth=str(depth)
    self.__uri=uri
    self.__has_body=None    # did we parse a body?

    def read_propfind(self,xml_doc):
        self.request_type,self.proplist=utils.parse_propfind(xml_doc)

    def createResponse(self,*args):
        """ create the multistatus response

    This will be delegated to the specific method
    depending on which request (allprop, propname, prop)
    was found.

    If we get a PROPNAME then we simply return the list with empty
    values which we get from the interface class

    If we get an ALLPROP we first get the list of properties and then
    we do the same as with a PROP method.

    """

    if self.request_type==RT_ALLPROP:
        return apply(self.create_allprop,args)
        if self.request_type==RT_PROPNAME:
        return apply(self.create_propname,args)
        if self.request_type==RT_PROP:
        return apply(self.create_prop,args)
        # no body means ALLPROP!
    return apply(self.create_allprop,args)


    def create_propname(self,*args):
        """ create a multistatus response for the prop names """
    dc=self.__dataclass
    # create the document generator
    doc = Document(None)
    ms=doc.createElement("D:multistatus")
    ms.setAttribute("xmlns:D","DAV:")
    doc.appendChild(ms)

    if self.__depth=="0":
        pnames=dc.get_propnames(self.__uri)
        re=self.mk_propname_response(self.__uri,pnames)
        ms.appendChild(re)

        elif self.__depth=="1":
        pnames=dc.get_propnames(self.__uri)
        re=self.mk_propname_response(self.__uri,pnames)
        ms.appendChild(re)
        for newuri in dc.get_childs(self.__uri):
            pnames=dc.get_propnames(newuri)
        re=self.mk_propname_response(newuri,pnames)
        ms.appendChild(re)
    # *** depth=="infinity"

    sfile=StringIO()
    ext.PrettyPrint(doc,stream=sfile)
    s=sfile.getvalue()
    sfile.close()
    return s

    def create_allprop(self,*args):
        """ return a list of all properties """
    self.proplist={}
    for ns,plist in apply(self.__dataclass.get_propnames,(self.__uri,)+args).items():
        self.proplist[ns]=plist
    return apply(self.create_prop,args)

    def create_prop(self,*args):
        """ handle a <prop> request

    This will

    1. set up the <multistatus>-Framework

    2. read the property values for each URI
    (which is dependant on the Depth header)
    This is done by the get_propvalues() method.

    3. For each URI call the append_result() method
    to append the actual <result>-Tag to the result
    document.

    We differ between "good" properties, which have been
    assigned a value by the interface class and "bad"
    properties, which resulted in an error, either 404
    (Not Found) or 403 (Forbidden).

    """


    # create the document generator
    doc = Domlette.Document()
    ms=doc.createElementNS(DAV_NS,"D:multistatus")
    doc.appendChild(ms)

    if self.__depth=="0":
        gp,bp=apply(self.get_propvalues,(self.__uri,)+args)
        res=self.mk_prop_response(self.__uri,gp,bp,doc)
        ms.appendChild(res)

        elif self.__depth=="1":
        gp,bp=apply(self.get_propvalues,(self.__uri,)+args)
        res=self.mk_prop_response(self.__uri,gp,bp,doc)
        ms.appendChild(res)
            children = apply(self.__dataclass.get_childs,(self.__uri,)+args)

        for newuri in children:
            gp,bp=apply(self.get_propvalues,(newuri,)+args)
        res=self.mk_prop_response(newuri,gp,bp,doc)
        ms.appendChild(res)

    sfile=StringIO()
    ext.PrettyPrint(doc,stream=sfile)
    s=sfile.getvalue()
    sfile.close()
    return s


    def mk_propname_response(self,uri,propnames,doc):
        """ make a new <prop> result element for a PROPNAME request

    This will simply format the propnames list.
    propnames should have the format {NS1 : [prop1, prop2, ...], NS2: ...}

    """
    re=doc.createElement("D:response")

    # write href information
    uparts=urlparse.urlparse(uri)
    fileloc=uparts[2]
    href=doc.createElement("D:href")
    huri=doc.createTextNode(urllib.quote(fileloc))
    href.appendChild(huri)
    re.appendChild(href)

    ps=doc.createElement("D:propstat")
    nsnum = 1

    for ns,plist in propnames.items():
        # write prop element
        pr=doc.createElement("D:prop")
        nsp="ns"+str(nsnum)
        pr.setAttribute("xmlns:"+nsp,ns)
        nsnum=nsnum+1

        # write propertynames
        for p in plist:
            pe=doc.createElement(nsp+":"+p)
        pr.appendChild(pe)
        ps.appendChild(pr)
    re.appendChild(ps)
    return re



    def mk_prop_response(self,uri,good_props,bad_props,doc):
        """ make a new <prop> result element

    We differ between the good props and the bad ones for
    each generating an extra <propstat>-Node (for each error
    one, that means).

    """
    re=doc.createElementNS(DAV_NS,"D:response")

    # write href information
    uparts=urlparse.urlparse(uri)
    fileloc=uparts[2]
    href=doc.createElementNS(DAV_NS,"D:href")
    huri=doc.createTextNode(urllib.quote(fileloc))
    href.appendChild(huri)
    re.appendChild(href)

    # write good properties
        nsMap = {DAV_NS:'D'}
        nsnum = 2
    if good_props:
        ps=doc.createElementNS(DAV_NS,"D:propstat")
        re.appendChild(ps)
        gp=doc.createElementNS(DAV_NS,"D:prop")
        for ns in good_props.keys():
                ns_prefix = nsMap.get(ns)
                if not ns_prefix:
                    ns_prefix="ns"+str(nsnum)
                    nsnum = nsnum + 1
                    nsMap[ns] = ns_prefix
        for p,v in good_props[ns].items():
            pe=doc.createElementNS(DAV_NS,ns_prefix+':' + str(p))
            if p=="resourcetype":
                if v=="1":
                ve=doc.createElementNS(DAV_NS,"D:collection")
                pe.appendChild(ve)
            else:
                ve=doc.createTextNode(str(v))
            pe.appendChild(ve)
                    gp.appendChild(pe)
        ps.appendChild(gp)
        s=doc.createElementNS(DAV_NS,"D:status")
        t=doc.createTextNode("HTTP/1.1 200 OK")
        s.appendChild(t)
        ps.appendChild(s)
        re.appendChild(ps)

    # now write the errors!
    if len(bad_props.items()):
        # write a propstat for each error code
        for ecode in bad_props.keys():
            ps=doc.createElementNS(DAV_NS,"D:propstat")
        re.appendChild(ps)
        bp=doc.createElementNS(DAV_NS,"D:prop")
        ps.appendChild(bp)
        for ns in bad_props[ecode].keys():
                    ns_prefix = nsMap.get(ns)
                    if not ns_prefix:
                        ns_prefix="ns"+str(nsnum)
                        nsnum = nsnum + 1
                        nsMap[ns] = ns_prefix
            for p in bad_props[ecode][ns]:
                pe=doc.createElementNS(ns,ns_prefix+':'+str(p))
            bp.appendChild(pe)
        s=doc.createElementNS(DAV_NS,"D:status")
        t=doc.createTextNode(utils.gen_estring(ecode))
        s.appendChild(t)
        ps.appendChild(s)
        re.appendChild(ps)


    # return the new response element
    return re


    def get_propvalues(self,uri,*args):
        """ create lists of property values for an URI

    We create two lists for an URI: the properties for
    which we found a value and the ones for which we
    only got an error, either because they haven't been
    found or the user is not allowed to read them.

    """
    good_props={}
    bad_props={}

    for (ns,plist) in self.proplist.items():
        good_props[ns]={}
        bad_props={}
        for prop in plist:
            try:
            r=apply(self.__dataclass.get_prop,(uri,ns,prop)+args)
            good_props[ns][prop]=str(r)
        except DAV_Error, error_code:
            ec=error_code[0]
            # ignore props with error_code if 0 (invisible)
            if ec==0: continue
            if bad_props.has_key(ec):
                if bad_props[ec].has_key(ns):
                bad_props[ec][ns].append(prop)
                else:
                bad_props[ec][ns]=[prop]
            else:
                bad_props[ec]={ns:[prop]}

    return good_props, bad_props

### vim: sts=4

